@@ -33,7 +33,7 @@
  *           files (until the tracker_miner_fs_file_notify() is called by the
  *           upper layer).
  *   1.2. "READY" tasks are those used to specify tasks which have a proper
- *         SPARQL string ready to be pushed to tracker-store. If
+ *         SPARQL string ready to be pushed to tracker-store.
  *   1.3. "PROCESSING" tasks are those used to specify tasks that are currently
  *         being pushed to the store.
  *
@@ -64,17 +64,16 @@
  * 3. The number of tasks pushed to the pull as "WAIT" tasks is limited to the
  *    number set while creating the pool. This value corresponds to the
  *    "processing-pool-wait-limit" property in the TrackerMinerFS object, and
- *    currently is set to 1 for TrackerMinerApplications and to 10 to
+ *    currently is set to 10 for both TrackerMinerApplications and
  *    TrackerMinerFiles. In the case of TrackerMinerFiles, this number specifies
  *    the maximum number of extraction requests that can be managed in parallel.
  *
  * 4. The number of tasks pushed to the pull as "READY" tasks is limited to
  *    the number set while creating the pool. This value corresponds to the
  *    "processing-pool-ready-limit" property in the TrackerMinerFS object, and
- *    currently is set to 1 for TrackerMinerApplications and to 100 to
- *    TrackerMinerFiles. In the case of TrackerMinerFiles, this number specifies
- *    the maximum number of SPARQL updates that can be merged into a single
- *    multi-insert SPARQL connection.
+ *    currently is set to 100 for both TrackerMinerApplications and
+ *    TrackerMinerFiles. This number specifies the maximum number of SPARQL
+ *    updates that can be merged into a single multi-insert SPARQL connection.
  *
  * 5. When a task is pushed to the pool as a "READY" task, the pool will be in
  *    charge of executing the SPARQL update into the store.
@@ -112,6 +111,8 @@
 
 /* If defined, will dump additional traces */
 #ifdef PROCESSING_POOL_ENABLE_TRACE
+#warning Processing pool traces are enabled
+#define POOL_STATUS_TRACE_TIMEOUT_SECS 10
 #define trace(message, ...) g_debug (message, ##__VA_ARGS__)
 #else
 #define trace(...)
@@ -128,12 +129,33 @@ typedef enum {
 	TRACKER_PROCESSING_TASK_STATUS_LAST
 } TrackerProcessingTaskStatus;
 
+typedef struct {
+	const gchar *bulk_operation;
+	GList *tasks;
+	gchar *sparql;
+} BulkOperationMerge;
+
+typedef enum {
+	CONTENT_NONE,
+	CONTENT_SPARQL_STRING,
+	CONTENT_SPARQL_BUILDER,
+	CONTENT_BULK_OPERATION
+} TaskContentType;
+
 struct _TrackerProcessingTask {
 	/* The file being processed */
 	GFile *file;
-	/* The FULL sparql to be updated in the store */
-	TrackerSparqlBuilder *sparql;
-	gchar *sparql_string;
+
+	TaskContentType content;
+
+	union {
+		TrackerSparqlBuilder *builder;
+		gchar *string;
+		struct {
+			const gchar *bulk_operation;
+			TrackerBulkMatchType match;
+		} bulk;
+	} data;
 
 	/* The context of the task */
 	gpointer context;
@@ -156,8 +178,8 @@ struct _TrackerProcessingTask {
 };
 
 struct _TrackerProcessingPool {
-	/* Connection to the Store */
-	TrackerSparqlConnection *connection;
+	/* Owner of the pool */
+	TrackerMinerFS *miner;
 
 	/* The tasks currently in WAIT or PROCESS status */
 	GQueue *tasks[TRACKER_PROCESSING_TASK_STATUS_LAST];
@@ -168,8 +190,20 @@ struct _TrackerProcessingPool {
 	GPtrArray *sparql_buffer;
 	GFile *sparql_buffer_current_parent;
 	time_t sparql_buffer_start_time;
+
+#ifdef PROCESSING_POOL_ENABLE_TRACE
+	/* Timeout to notify status of the queues, if traces
+	 * enabled only. */
+	guint timeout_id;
+#endif /* PROCESSING_POOL_ENABLE_TRACE */
 };
 
+typedef struct {
+	GPtrArray *tasks;
+	GArray *error_map;
+	guint n_bulk_operations;
+} UpdateArrayData;
+
 /*------------------- PROCESSING TASK ----------------------*/
 
 TrackerProcessingTask *
@@ -188,6 +222,20 @@ tracker_processing_task_new (GFile *file)
 	return task;
 }
 
+static void
+tracker_processing_task_data_unset (TrackerProcessingTask *task)
+{
+	if (task->content == CONTENT_SPARQL_STRING) {
+		g_free (task->data.string);
+	} else if (task->content == CONTENT_SPARQL_BUILDER) {
+		if (task->data.builder) {
+			g_object_unref (task->data.builder);
+		}
+	}
+
+	task->content = CONTENT_NONE;
+}
+
 void
 tracker_processing_task_free (TrackerProcessingTask *task)
 {
@@ -203,10 +251,9 @@ tracker_processing_task_free (TrackerProcessingTask *task)
 	    task->context_free_func) {
 		task->context_free_func (task->context);
 	}
-	if (task->sparql) {
-		g_object_unref (task->sparql);
-	}
-	g_free (task->sparql_string);
+
+	tracker_processing_task_data_unset (task);
+
 	g_object_unref (task->file);
 	g_slice_free (TrackerProcessingTask, task);
 }
@@ -242,34 +289,82 @@ void
 tracker_processing_task_set_sparql (TrackerProcessingTask *task,
                                     TrackerSparqlBuilder  *sparql)
 {
-	if (task->sparql) {
-		g_object_unref (task->sparql);
+	tracker_processing_task_data_unset (task);
+
+	if (sparql) {
+		task->data.builder = g_object_ref (sparql);
+		task->content = CONTENT_SPARQL_BUILDER;
 	}
-	if (task->sparql_string) {
-		g_free (task->sparql_string);
-		task->sparql_string = NULL;
-	}
-	task->sparql = g_object_ref (sparql);
 }
 
 void
 tracker_processing_task_set_sparql_string (TrackerProcessingTask *task,
                                            gchar                 *sparql_string)
 {
-	if (task->sparql) {
-		g_object_unref (task->sparql);
-		task->sparql = NULL;
+	tracker_processing_task_data_unset (task);
+
+	if (sparql_string) {
+		/* We take ownership of the input string! */
+		task->data.string = sparql_string;
+		task->content = CONTENT_SPARQL_STRING;
 	}
-	if (task->sparql_string) {
-		g_free (task->sparql_string);
-	}
-	/* We take ownership of the input string! */
-	task->sparql_string = sparql_string;
 }
 
+void
+tracker_processing_task_set_bulk_operation (TrackerProcessingTask *task,
+                                            const gchar           *sparql,
+                                            TrackerBulkMatchType   match)
+{
+	tracker_processing_task_data_unset (task);
+
+	if (sparql) {
+		/* This string is expected to remain constant */
+		task->data.bulk.bulk_operation = sparql;
+		task->data.bulk.match = match;
+		task->content = CONTENT_BULK_OPERATION;
+	}
+}
 
 /*------------------- PROCESSING POOL ----------------------*/
 
+#ifdef PROCESSING_POOL_ENABLE_TRACE
+static const gchar *queue_names [TRACKER_PROCESSING_TASK_STATUS_LAST] = {
+	"WAIT",
+	"READY",
+	"PROCESSING"
+};
+
+static gboolean
+pool_status_trace_timeout_cb (gpointer data)
+{
+	TrackerProcessingPool *pool = data;
+	guint i;
+
+	trace ("(Processing Pool %s) ------------",
+	       G_OBJECT_TYPE_NAME (pool->miner));
+	for (i = TRACKER_PROCESSING_TASK_STATUS_WAIT;
+	     i < TRACKER_PROCESSING_TASK_STATUS_LAST;
+	     i++) {
+		GList *l;
+
+		l = g_queue_peek_head_link (pool->tasks[i]);
+		trace ("(Processing Pool %s) Queue %s has %u tasks",
+		       G_OBJECT_TYPE_NAME (pool->miner),
+		       queue_names[i],
+		       g_list_length (l));
+		while (l) {
+			trace ("(Processing Pool %s)     Task %p in queue %s",
+			       G_OBJECT_TYPE_NAME (pool->miner),
+			       l->data,
+			       queue_names[i]);
+			l = g_list_next (l);
+		}
+	}
+	return TRUE;
+}
+
+#endif /* PROCESSING_POOL_ENABLE_TRACE */
+
 static void
 pool_queue_free_foreach (gpointer data,
                          gpointer user_data)
@@ -291,6 +386,11 @@ tracker_processing_pool_free (TrackerProcessingPool *pool)
 	if (!pool)
 		return;
 
+#ifdef PROCESSING_POOL_ENABLE_TRACE
+	if (pool->timeout_id)
+		g_source_remove (pool->timeout_id);
+#endif /* PROCESSING_POOL_ENABLE_TRACE */
+
 	/* Free any pending task here... shouldn't really
 	 * be any */
 	for (i = TRACKER_PROCESSING_TASK_STATUS_WAIT;
@@ -310,20 +410,19 @@ tracker_processing_pool_free (TrackerProcessingPool *pool)
 		g_ptr_array_free (pool->sparql_buffer, TRUE);
 	}
 
-	g_object_unref (pool->connection);
 	g_free (pool);
 }
 
 TrackerProcessingPool *
-tracker_processing_pool_new (TrackerSparqlConnection *connection,
-                             guint                    limit_wait,
-                             guint                    limit_ready)
+tracker_processing_pool_new (TrackerMinerFS *miner,
+                             guint           limit_wait,
+                             guint           limit_ready)
 {
 	TrackerProcessingPool *pool;
 
 	pool = g_new0 (TrackerProcessingPool, 1);
 
-	pool->connection = g_object_ref (connection);
+	pool->miner = miner;
 	pool->limit[TRACKER_PROCESSING_TASK_STATUS_WAIT] = limit_wait;
 	pool->limit[TRACKER_PROCESSING_TASK_STATUS_READY] = limit_ready;
 	/* convenience limit, not really used currently */
@@ -339,6 +438,12 @@ tracker_processing_pool_new (TrackerSparqlConnection *connection,
 	         limit_wait,
 	         limit_ready);
 
+#ifdef PROCESSING_POOL_ENABLE_TRACE
+	pool->timeout_id = g_timeout_add_seconds (POOL_STATUS_TRACE_TIMEOUT_SECS,
+	                                          pool_status_trace_timeout_cb,
+	                                          pool);
+#endif /* PROCESSING_POOL_ENABLE_TRACE */
+
 	return pool;
 }
 
@@ -346,7 +451,8 @@ void
 tracker_processing_pool_set_wait_limit (TrackerProcessingPool *pool,
                                         guint                  limit)
 {
-	g_message ("Processing pool limit for WAIT tasks set to %u", limit);
+	g_message ("Processing pool limit for WAIT tasks set to %u",
+	           limit);
 	pool->limit[TRACKER_PROCESSING_TASK_STATUS_WAIT] = limit;
 }
 
@@ -354,7 +460,8 @@ void
 tracker_processing_pool_set_ready_limit (TrackerProcessingPool *pool,
                                          guint                  limit)
 {
-	g_message ("Processing pool limit for READY tasks set to %u", limit);
+	g_message ("Processing pool limit for READY tasks set to %u",
+	           limit);
 	pool->limit[TRACKER_PROCESSING_TASK_STATUS_READY] = limit;
 }
 
@@ -436,8 +543,10 @@ tracker_processing_pool_push_wait_task (TrackerProcessingPool *pool,
 	task->status = TRACKER_PROCESSING_TASK_STATUS_WAIT;
 
 
-	trace ("(Processing Pool) Pushed WAIT task %p for file '%s'",
-	       task, task->file_uri);
+	trace ("(Processing Pool %s) Pushed WAIT task %p for file '%s'",
+	       G_OBJECT_TYPE_NAME (pool->miner),
+	       task,
+	       task->file_uri);
 
 	/* Push a new task in WAIT status (so just add it to the tasks queue,
 	 * and don't process it. */
@@ -458,7 +567,8 @@ tracker_processing_pool_sparql_update_cb (GObject      *object,
 	task = user_data;
 
 	trace ("(Processing Pool) Finished update of task %p for file '%s'",
-	       task, task->file_uri);
+	       task,
+	       task->file_uri);
 
 	/* Before calling user-provided callback, REMOVE the task from the pool;
 	 * as the user-provided callback may actually modify the pool again */
@@ -479,66 +589,193 @@ tracker_processing_pool_sparql_update_array_cb (GObject      *object,
 {
 	GError *global_error = NULL;
 	GPtrArray *sparql_array_errors;
-	GPtrArray *sparql_array;
+	UpdateArrayData *update_data;
 	guint i;
 
 	/* Get arrays of errors and queries */
-	sparql_array = user_data;
+	update_data = user_data;
 
 	trace ("(Processing Pool) Finished array-update of tasks %p",
-	       sparql_array);
+	       update_data->tasks);
 
 	sparql_array_errors = tracker_sparql_connection_update_array_finish (TRACKER_SPARQL_CONNECTION (object),
 	                                                                     result,
 	                                                                     &global_error);
 	if (global_error) {
-		g_critical ("(Sparql buffer) Could not execute array-update of tasks %p with '%u' items: %s",
-		            sparql_array,
-		            sparql_array->len,
+		g_critical ("(Processing Pool) Could not execute array-update of tasks %p with '%u' items: %s",
+		            update_data->tasks,
+		            update_data->tasks->len,
 		            global_error->message);
 	}
 
 	/* Report status on each task of the batch update */
-	for (i = 0; i < sparql_array->len; i++) {
+	for (i = 0; i < update_data->tasks->len; i++) {
 		TrackerProcessingTask *task;
+		GError *error = NULL;
 
-		task = g_ptr_array_index (sparql_array, i);
+		task = g_ptr_array_index (update_data->tasks, i);
 
 		/* Before calling user-provided callback, REMOVE the task from the pool;
 		 * as the user-provided callback may actually modify the pool again */
 		tracker_processing_pool_remove_task (task->pool, task);
 
+		if (global_error) {
+			error = global_error;
+		} else {
+			gint error_pos;
+
+			error_pos = g_array_index (update_data->error_map, gint, i);
+
+			/* Find the corresponing error according to the passed map,
+			 * numbers >= 0 are non-bulk tasks, and < 0 are bulk tasks,
+			 * so the number of bulk operations must be added, as these
+			 * tasks are prepended.
+			 */
+			error_pos += update_data->n_bulk_operations;
+			error = g_ptr_array_index (sparql_array_errors, error_pos);
+		}
+
 		/* Call finished handler with the error, if any */
-		task->finished_handler (task, task->finished_user_data,
-		                        (global_error ?
-		                         global_error :
-		                         g_ptr_array_index (sparql_array_errors, i)));
+		task->finished_handler (task, task->finished_user_data, error);
 
 		/* No need to deallocate the task here, it will be done when
-		 * unref-ing the GPtrArray below */
+		 * unref-ing the UpdateArrayData below */
 	}
 
 	/* Unref the arrays of errors and queries */
 	if (sparql_array_errors)
 		g_ptr_array_unref (sparql_array_errors);
+
 	/* Note that tasks are actually deallocated here */
-	g_ptr_array_unref (sparql_array);
-	g_clear_error (&global_error);
+	g_ptr_array_free (update_data->tasks, TRUE);
+	g_array_free (update_data->error_map, TRUE);
+	g_slice_free (UpdateArrayData, update_data);
+
+	if (global_error) {
+		g_error_free (global_error);
+	}
+}
+
+static void
+bulk_operation_merge_finish (BulkOperationMerge *merge)
+{
+	if (merge->sparql) {
+		g_free (merge->sparql);
+		merge->sparql = NULL;
+	}
+
+	if (merge->bulk_operation && merge->tasks) {
+		GString *equals_string = NULL, *children_string = NULL, *sparql;
+		guint n_equals = 0;
+		GList *l;
+
+		for (l = merge->tasks; l; l = l->next) {
+			TrackerProcessingTask *task = l->data;
+			gchar *uri;
+
+			uri = g_file_get_uri (task->file);
+
+			if (task->data.bulk.match & TRACKER_BULK_MATCH_EQUALS) {
+				if (!equals_string) {
+					equals_string = g_string_new ("");
+				} else {
+					g_string_append_c (equals_string, ',');
+				}
+
+				g_string_append_printf (equals_string, "\"%s\"", uri);
+				n_equals++;
+			}
+
+			if (task->data.bulk.match & TRACKER_BULK_MATCH_CHILDREN) {
+				if (!children_string) {
+					children_string = g_string_new ("");
+				} else {
+					g_string_append_c (children_string, ',');
+				}
+
+				g_string_append_printf (children_string, "\"%s\"", uri);
+			}
+
+			g_free (uri);
+		}
+
+		sparql = g_string_new ("");
+
+		if (equals_string) {
+			g_string_append (sparql, merge->bulk_operation);
+
+			if (n_equals == 1) {
+				g_string_append_printf (sparql,
+				                        " WHERE { "
+				                        "  ?f nie:url %s"
+				                        "} ",
+				                        equals_string->str);
+			} else {
+				g_string_append_printf (sparql,
+				                        " WHERE { "
+				                        "  ?f nie:url ?u ."
+				                        "  FILTER (?u IN (%s))"
+				                        "} ",
+				                        equals_string->str);
+			}
+
+			g_string_free (equals_string, TRUE);
+		}
+
+		if (children_string) {
+			g_string_append (sparql, merge->bulk_operation);
+			g_string_append_printf (sparql,
+			                        " WHERE { "
+			                        "  ?f nie:url ?u ."
+			                        "  FILTER (tracker:uri-is-descendant (%s, ?u))"
+			                        "} ",
+			                        children_string->str);
+			g_string_free (children_string, TRUE);
+		}
+
+		merge->sparql = g_string_free (sparql, FALSE);
+	}
+}
+
+static BulkOperationMerge *
+bulk_operation_merge_new (const gchar *bulk_operation)
+{
+	BulkOperationMerge *operation;
+
+	operation = g_slice_new0 (BulkOperationMerge);
+	operation->bulk_operation = bulk_operation;
+
+	return operation;
+}
+
+static void
+bulk_operation_merge_free (BulkOperationMerge *operation)
+{
+	g_list_free (operation->tasks);
+	g_free (operation->sparql);
+	g_slice_free (BulkOperationMerge, operation);
 }
 
 void
-tracker_processing_pool_buffer_flush (TrackerProcessingPool *pool)
+tracker_processing_pool_buffer_flush (TrackerProcessingPool *pool,
+                                      const gchar           *reason)
 {
-	guint i;
-	gchar **sparql_array;
+	GPtrArray *bulk_ops;
+	GArray *sparql_array, *error_map;
+	UpdateArrayData *update_data;
+	guint i, j;
 
 	if (!pool->sparql_buffer)
 		return;
 
 	/* Loop buffer and construct array of strings */
-	sparql_array = g_new (gchar *, pool->sparql_buffer->len);
+	sparql_array = g_array_new (FALSE, TRUE, sizeof (gchar *));
+	bulk_ops = g_ptr_array_new_with_free_func ((GDestroyNotify) bulk_operation_merge_free);
+	error_map = g_array_new (TRUE, TRUE, sizeof (gint));
+
 	for (i = 0; i < pool->sparql_buffer->len; i++) {
 		TrackerProcessingTask *task;
+		gint pos;
 
 		task = g_ptr_array_index (pool->sparql_buffer, i);
 
@@ -552,22 +789,70 @@ tracker_processing_pool_buffer_flush (TrackerProcessingPool *pool)
 		task->pool = pool;
 		g_queue_push_head (pool->tasks[TRACKER_PROCESSING_TASK_STATUS_PROCESSING], task);
 
-		/* Add original string, not a duplicate */
-		sparql_array[i] = (task->sparql ?
-		                   (gchar *) tracker_sparql_builder_get_result (task->sparql) :
-		                   task->sparql_string);
+		if (task->content == CONTENT_SPARQL_STRING) {
+			g_array_append_val (sparql_array, task->data.string);
+			pos = sparql_array->len - 1;
+		} else if (task->content == CONTENT_SPARQL_BUILDER) {
+			const gchar *str = tracker_sparql_builder_get_result (task->data.builder);
+			g_array_append_val (sparql_array, str);
+			pos = sparql_array->len - 1;
+		} else if (task->content == CONTENT_BULK_OPERATION) {
+			BulkOperationMerge *bulk = NULL;
+			gint j;
+
+			for (j = 0; j < bulk_ops->len; j++) {
+				BulkOperationMerge *cur;
+
+				cur = g_ptr_array_index (bulk_ops, j);
+
+				if (cur->bulk_operation == task->data.bulk.bulk_operation) {
+					bulk = cur;
+					pos = - 1 - j;
+					break;
+				}
+			}
+
+			if (!bulk) {
+				bulk = bulk_operation_merge_new (task->data.bulk.bulk_operation);
+				g_ptr_array_add (bulk_ops, bulk);
+				pos = - bulk_ops->len;
+			}
+
+			bulk->tasks = g_list_prepend (bulk->tasks, task);
+		}
+
+		g_array_append_val (error_map, pos);
 	}
 
-	trace ("(Processing Pool) Flushing array-update of tasks %p with %u items",
-	       pool->sparql_buffer, pool->sparql_buffer->len);
+	for (j = 0; j < bulk_ops->len; j++) {
+		BulkOperationMerge *bulk;
 
-	tracker_sparql_connection_update_array_async (pool->connection,
-	                                              sparql_array,
-	                                              pool->sparql_buffer->len,
+		bulk = g_ptr_array_index (bulk_ops, j);
+		bulk_operation_merge_finish (bulk);
+
+		if (bulk->sparql) {
+			g_array_prepend_val (sparql_array, bulk->sparql);
+		}
+	}
+
+	trace ("(Processing Pool %s) Flushing array-update of tasks %p with %u items (%s)",
+	       G_OBJECT_TYPE_NAME (pool->miner),
+	       pool->sparql_buffer,
+	       pool->sparql_buffer->len,
+	       reason ? reason : "Unknown reason");
+
+	update_data = g_slice_new0 (UpdateArrayData);
+	update_data->tasks = pool->sparql_buffer;
+	update_data->n_bulk_operations = bulk_ops->len;
+	update_data->error_map = error_map;
+
+	tracker_sparql_connection_update_array_async (tracker_miner_get_connection (TRACKER_MINER (pool->miner)),
+	                                              (gchar **) sparql_array->data,
+	                                              sparql_array->len,
 	                                              G_PRIORITY_DEFAULT,
 	                                              NULL,
 	                                              tracker_processing_pool_sparql_update_array_cb,
-	                                              pool->sparql_buffer);
+	                                              update_data);
 
 	/* Clear current parent */
 	if (pool->sparql_buffer_current_parent) {
@@ -576,7 +861,10 @@ tracker_processing_pool_buffer_flush (TrackerProcessingPool *pool)
 	}
 
 	/* Clear temp buffer */
-	g_free (sparql_array);
+	g_array_free (sparql_array, TRUE);
+
+	g_ptr_array_free (bulk_ops, TRUE);
+
 	pool->sparql_buffer_start_time = 0;
 	/* Note the whole buffer is passed to the update_array callback,
 	 * so no need to free it. */
@@ -592,8 +880,8 @@ tracker_processing_pool_push_ready_task (TrackerProcessingPool
 {
 	GList *previous;
 
-	/* The task MUST have a proper SPARQL here */
-	g_assert (task->sparql != NULL || task->sparql_string != NULL);
+	/* The task MUST have a proper content here */
+	g_assert (task->content != CONTENT_NONE);
 
 	/* First, check if the task was already added as being WAITING */
 	previous = g_queue_find (pool->tasks[TRACKER_PROCESSING_TASK_STATUS_WAIT], task);
@@ -613,25 +901,53 @@ tracker_processing_pool_push_ready_task (TrackerProcessingPool
 	/* If buffering not requested, OR the limit of READY tasks is actually 1,
 	 * flush previous buffer (if any) and then the new update */
 	if (!buffer || pool->limit[TRACKER_PROCESSING_TASK_STATUS_READY] == 1) {
-		trace ("(Processing Pool) Pushed READY/PROCESSING task %p for file '%s'",
-		       task, task->file_uri);
+		BulkOperationMerge *operation = NULL;
+		const gchar *sparql = NULL;
+
+		trace ("(Processing Pool %s) Pushed READY/PROCESSING task %p for file '%s'",
+		       G_OBJECT_TYPE_NAME (pool->miner),
+		       task,
+		       task->file_uri);
 
 		/* Flush previous */
-		tracker_processing_pool_buffer_flush (pool);
+		tracker_processing_pool_buffer_flush (pool,
+		                                      "Before unbuffered task");
 
 		/* Set status of the task as PROCESSING (No READY status here!) */
 		task->status = TRACKER_PROCESSING_TASK_STATUS_PROCESSING;
 		g_queue_push_head (pool->tasks[TRACKER_PROCESSING_TASK_STATUS_PROCESSING], task);
 
+		trace ("(Processing Pool %s) Flushing single task %p",
+		       G_OBJECT_TYPE_NAME (pool->miner),
+		       task);
+
 		/* And update the new one */
-		tracker_sparql_connection_update_async (pool->connection,
-		                                        (task->sparql ?
-		                                         tracker_sparql_builder_get_result (task->sparql) :
-		                                         task->sparql_string),
-		                                        G_PRIORITY_DEFAULT,
-		                                        NULL,
-		                                        tracker_processing_pool_sparql_update_cb,
-		                                        task);
+		if (task->content == CONTENT_SPARQL_STRING) {
+			sparql = task->data.string;
+		} else if (task->content == CONTENT_SPARQL_BUILDER) {
+			sparql = tracker_sparql_builder_get_result (task->data.builder);
+		} else if (task->content == CONTENT_BULK_OPERATION) {
+			operation = bulk_operation_merge_new (task->data.bulk.bulk_operation);
+			operation->tasks = g_list_prepend (NULL, task);
+			bulk_operation_merge_finish (operation);
+
+			if (operation->sparql) {
+				sparql = operation->sparql;
+			}
+		}
+
+		if (sparql) {
+			tracker_sparql_connection_update_async (tracker_miner_get_connection (TRACKER_MINER (pool->miner)),
+			                                        sparql,
+			                                        G_PRIORITY_DEFAULT,
+			                                        NULL,
+			                                        tracker_processing_pool_sparql_update_cb,
+			                                        task);
+		}
+
+		if (operation) {
+			bulk_operation_merge_free (operation);
+		}
 
 		return TRUE;
 	} else {
@@ -657,8 +973,11 @@ tracker_processing_pool_push_ready_task (TrackerProcessingPool
 			pool->sparql_buffer_current_parent = g_object_ref (parent);
 		}
 
-		trace ("(Processing Pool) Pushed READY task %p for file '%s' into array %p",
-		       task, task->file_uri, pool->sparql_buffer);
+		trace ("(Processing Pool %s) Pushed READY task %p for file '%s' into array %p",
+		       G_OBJECT_TYPE_NAME (pool->miner),
+		       task,
+		       task->file_uri,
+		       pool->sparql_buffer);
 
 		/* Add task to array */
 		g_ptr_array_add (pool->sparql_buffer, task);
@@ -669,12 +988,21 @@ tracker_processing_pool_push_ready_task (TrackerProcessingPool
 		 *  - Maximum number of READY items reached
 		 *  - Not flushed in the last MAX_SPARQL_BUFFER_TIME seconds
 		 */
-		if (!parent ||
-		    !g_file_equal (parent, pool->sparql_buffer_current_parent) ||
-		    tracker_processing_pool_ready_limit_reached (pool) ||
-		    (time (NULL) - pool->sparql_buffer_start_time > MAX_SPARQL_BUFFER_TIME)) {
-			/* Flush! */
-			tracker_processing_pool_buffer_flush (pool);
+		if (!parent) {
+			tracker_processing_pool_buffer_flush (pool,
+			                                      "File with no parent");
+			flushed = TRUE;
+		} else if (!g_file_equal (parent, pool->sparql_buffer_current_parent)) {
+			tracker_processing_pool_buffer_flush (pool,
+			                                      "Different parent");
+			flushed = TRUE;
+		} else if (tracker_processing_pool_ready_limit_reached (pool)) {
+			tracker_processing_pool_buffer_flush (pool,
+			                                      "Ready limit reached");
+			flushed = TRUE;
+		} else if (time (NULL) - pool->sparql_buffer_start_time > MAX_SPARQL_BUFFER_TIME) {
+			tracker_processing_pool_buffer_flush (pool,
+			                                      "Buffer time reached");
 			flushed = TRUE;
 		}
 
