@@ -43,10 +43,8 @@
 #include <libtracker-miner/tracker-miner.h>
 
 #include <libtracker-data/tracker-db-manager.h>
-#include <libtracker-data/tracker-db-dbus.h>
 
 #include "tracker-config.h"
-#include "tracker-dbus.h"
 #include "tracker-marshal.h"
 #include "tracker-miner-applications.h"
 #include "tracker-miner-files.h"
@@ -190,23 +188,32 @@ initialize_priority (void)
 }
 
 static gboolean
-should_crawl (TrackerConfig *config)
+should_crawl (TrackerConfig *config,
+              gboolean      *forced)
 {
 	gint crawling_interval;
 
 	crawling_interval = tracker_config_get_crawling_interval (config);
 
-	g_message ("Checking whether to perform mtime checks during crawling:");
+	g_message ("Checking whether to crawl file system based on configured crawling interval:");
 
-	if (crawling_interval == -1) {
+	if (crawling_interval == -2) {
 		g_message ("  Disabled");
 		return FALSE;
+	} else if (crawling_interval == -1) {
+		g_message ("  Maybe (depends on a clean last shutdown)");
+		return TRUE;
 	} else if (crawling_interval == 0) {
-		g_message ("  Enabled");
+		g_message ("  Forced");
+
+		if (forced) {
+			*forced = TRUE;
+		}
+
 		return TRUE;
 	} else {
 		guint64 then, now;
-		
+
 		then = tracker_db_manager_get_last_crawl_done ();
 
 		if (then < 1) {
@@ -261,10 +268,10 @@ miner_finished_cb (TrackerMinerFS *fs,
                    guint           total_files_ignored,
                    gpointer        user_data)
 {
-	g_message ("Finished mining in seconds:%f, total directories:%d, total files:%d",
-	           seconds_elapsed,
-	           total_directories_found,
-	           total_files_found);
+	tracker_info ("Finished mining in seconds:%f, total directories:%d, total files:%d",
+	              seconds_elapsed,
+	              total_directories_found,
+	              total_files_found);
 
 	if (TRACKER_IS_MINER_FILES (fs) &&
 	    tracker_miner_fs_get_initial_crawling (fs)) {
@@ -531,10 +538,13 @@ main (gint argc, gchar *argv[])
 {
 	TrackerConfig *config;
 	TrackerMiner *miner_applications, *miner_files;
-	TrackerMinerFilesIndex *object;
+	TrackerMinerFilesIndex *miner_files_index;
 	GOptionContext *context;
 	GError *error = NULL;
 	gchar *log_filename = NULL;
+	gboolean do_mtime_checking;
+	gboolean do_crawling;
+	gboolean force_mtime_checking = FALSE;
 
 	g_type_init ();
 
@@ -600,77 +610,122 @@ main (gint argc, gchar *argv[])
 
 	sanity_check_option_values (config);
 
-	/* Make sure we initialize DBus, this shows we are started
-	 * successfully when called upon from the daemon.
-	 */
-	if (!tracker_dbus_init ()) {
-		g_object_unref (config);
-		tracker_log_shutdown ();
-
-		return EXIT_FAILURE;
-	}
-
 	main_loop = g_main_loop_new (NULL, FALSE);
 
-	miner_files = tracker_miner_files_new (config);
-	tracker_miner_fs_set_initial_crawling (TRACKER_MINER_FS (miner_files),
-	                                       should_crawl (config));
+	/* Check if we should crawl and if we should force mtime
+	 * checking based on the config.
+	 */
+	do_crawling = should_crawl (config, &force_mtime_checking);
 
+	/* Get the last shutdown state to see if we need to perform a
+	 * full mtime check against the db or not.
+	 *
+	 * Set to TRUE here in case we crash and miss file system
+	 * events.
+	 */
+	g_message ("Checking whether to force mtime checking during crawling (based on last clean shutdown):");
+
+	/* Override the shutdown state decision based on the config */
+	if (force_mtime_checking) {
+		do_mtime_checking = TRUE;
+	} else {
+		do_mtime_checking = tracker_db_manager_get_need_mtime_check ();
+	}
+
+	g_message ("  %s %s",
+	           do_mtime_checking ? "Yes" : "No",
+	           force_mtime_checking ? "(forced from config)" : "");
+
+	/* Set the need for an mtime check to TRUE so we check in the
+	 * event of a crash, this is changed back on shutdown if
+	 * everything appears to be fine.
+	 */
+	tracker_db_manager_set_need_mtime_check (TRUE);
+
+	/* Create new TrackerMinerFiles object */
+	miner_files = tracker_miner_files_new (config, &error);
+	if (!miner_files) {
+		g_critical ("Couldn't create new Files miner: '%s'",
+		            error ? error->message : "unknown error");
+		g_object_unref (config);
+		tracker_log_shutdown ();
+		return EXIT_FAILURE;
+	}
+	tracker_miner_fs_set_initial_crawling (TRACKER_MINER_FS (miner_files), do_crawling);
+	tracker_miner_fs_set_mtime_checking (TRACKER_MINER_FS (miner_files), do_mtime_checking);
 	g_signal_connect (miner_files, "finished",
 			  G_CALLBACK (miner_finished_cb),
 			  NULL);
 
-	object = tracker_miner_files_index_new (TRACKER_MINER_FILES (miner_files));
-
-	if (!object) {
+	/* Create new TrackerMinerFilesIndex object */
+	miner_files_index = tracker_miner_files_index_new (TRACKER_MINER_FILES (miner_files));
+	if (!miner_files_index) {
 		g_object_unref (miner_files);
 		g_object_unref (config);
-		tracker_dbus_shutdown ();
 		tracker_log_shutdown ();
-
-		return EXIT_FAILURE;
-	}
-
-	/* Make Tracker available for introspection */
-	if (!tracker_dbus_register_objects (object)) {
-		g_object_unref (miner_files);
-		g_object_unref (object);
-		g_object_unref (config);
-		tracker_dbus_shutdown ();
-		tracker_log_shutdown ();
-
 		return EXIT_FAILURE;
 	}
 
 	/* Create miner for applications */
-	miner_applications = tracker_miner_applications_new ();
-	miners = g_slist_append (miners, miner_applications);
+	miner_applications = tracker_miner_applications_new (&error);
+	if (!miner_applications) {
+		g_critical ("Couldn't create new Applications miner: '%s'",
+		            error ? error->message : "unknown error");
+		g_object_unref (miner_files_index);
+		g_object_unref (miner_files);
+		g_object_unref (config);
+		tracker_log_shutdown ();
+		return EXIT_FAILURE;
+	}
+	tracker_miner_fs_set_initial_crawling (TRACKER_MINER_FS (miner_applications), do_crawling);
+
+	/* If a locale change was detected, always do mtime checks */
+	if (tracker_miner_applications_detect_locale_changed (miner_applications)) {
+		if (!do_mtime_checking)
+			g_debug ("Forcing mtime check in applications miner as locale change was detected");
+		tracker_miner_fs_set_mtime_checking (TRACKER_MINER_FS (miner_applications), TRUE);
+	} else {
+		tracker_miner_fs_set_mtime_checking (TRACKER_MINER_FS (miner_applications), do_mtime_checking);
+	}
 
 	g_signal_connect (miner_applications, "finished",
 	                  G_CALLBACK (miner_finished_cb),
 	                  NULL);
 
-	/* Create miner for files */
-	miners = g_slist_append (miners, miner_files);
+	/* Setup miners, applications first in list */
+	miners = g_slist_prepend (miners, miner_files);
+	miners = g_slist_prepend (miners, miner_applications);
 
 	tracker_thumbnailer_init ();
 
 	miner_handle_next ();
 
+	/* Go, go, go! */
 	g_main_loop_run (main_loop);
 
 	g_message ("Shutdown started");
 
+	/* Reasons to not mark ourselves as cleanly shutdown include:
+	 *
+	 * 1. Still have files to process in our queues.
+	 * 2. We crash (out of our control usually anyway).
+	 * 3. At least one of the miners is PAUSED.
+	 */
+	if (!tracker_miner_fs_has_items_to_process (TRACKER_MINER_FS (miner_files)) &&
+	    !tracker_miner_is_paused (miner_applications) &&
+	    !tracker_miner_is_paused (miner_files)) {
+		tracker_db_manager_set_need_mtime_check (FALSE);
+	}
+
 	g_main_loop_unref (main_loop);
 	g_object_unref (config);
+	g_object_unref (miner_files_index);
 
 	tracker_thumbnailer_shutdown ();
 
 	g_slist_foreach (miners, (GFunc) finalize_miner, NULL);
 	g_slist_free (miners);
 
-	tracker_dbus_shutdown ();
-
 	tracker_log_shutdown ();
 
 	g_print ("\nOK\n\n");
